#include "Win32DxInterface.h"

#include "FileLogger.h"

#include <dxgi.h>
#include <dxgi1_6.h>
#include <d3d11.h>
#include <d3d11on12.h>
#include <d3d12.h>
#include <d3d12sdklayers.h>

#include <d2d1_3.h>
#include <d2d1_1.h>

#include <dwrite.h>

void Win32DxInterface::getHardwareAdapter(IDXGIAdapter1** ppAdapter)
{
	*ppAdapter = nullptr;
	for (UINT adapterIndex = 0; ; ++adapterIndex)
	{
		IDXGIAdapter1* pAdapter = nullptr;
		if (mDxgiFactory->EnumAdapters1(adapterIndex, &pAdapter) == DXGI_ERROR_NOT_FOUND)
		{
			break;
		}
		if (SUCCEEDED(::D3D12CreateDevice(pAdapter, D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), nullptr)))
		{
			*ppAdapter = pAdapter;
			return;
		}
		pAdapter->Release();
	}
}

bool Win32DxInterface::isValid() const
{
	return mIsValid;
}

ID3D12Device* Win32DxInterface::getD3dDevice() const
{
	return mD3dDevice.Get();
}

ID3D11On12Device* Win32DxInterface::get11On12Device() const
{
	return mD3d11On12Device.Get();
}

IDXGIFactory7* Win32DxInterface::getDxgiFactory() const
{
	return mDxgiFactory.Get();
}

ID2D1Factory3* Win32DxInterface::getD2dFactory() const
{
	return mD2dFactory.Get();
}

IDWriteFactory* Win32DxInterface::getDirectWriteFactory() const
{
	return mDWriteFactory.Get();
}

ID2D1DeviceContext2* Win32DxInterface::getD2dContext() const
{
	return mD2dDeviceContext.Get();
}

ID3D12CommandQueue* Win32DxInterface::getCommandQueue() const
{
	return mCommandQueue.Get();
}

ID3D11DeviceContext* Win32DxInterface::getD3d11DeviceContext() const
{
	return mD3d11DeviceContext.Get();
}

void Win32DxInterface::initialize()
{
	MLOG_INFO("Initialize DirectX");
	createD3dFactory();

	Microsoft::WRL::ComPtr<IDXGIAdapter1> pAdapter;
	getHardwareAdapter(&pAdapter);
	createD3dDevice(pAdapter.Get());

	createCommandQueue();

	mIsValid = initializeD2d();
}

bool Win32DxInterface::createD3dFactory()
{
	UINT dxgiFactoryFlags = 0;
#if defined(DEBUG) || defined(_DEBUG)
	{
		Microsoft::WRL::ComPtr<ID3D12Debug> debugController;
		if (SUCCEEDED(::D3D12GetDebugInterface(IID_PPV_ARGS(&debugController))))
		{
			debugController->EnableDebugLayer();
			dxgiFactoryFlags |= DXGI_CREATE_FACTORY_DEBUG;
		}
	}
#endif
	if (FAILED(::CreateDXGIFactory2(dxgiFactoryFlags, IID_PPV_ARGS(&mDxgiFactory))))
	{
		MLOG_ERROR("Failed to create DXGI Factory");
	}
	return true;
}

bool Win32DxInterface::createD3dDevice(IDXGIAdapter1* pAdapter)
{
	if (pAdapter)
	{
		if (FAILED(::D3D12CreateDevice(pAdapter, D3D_FEATURE_LEVEL_11_0, IID_PPV_ARGS(&mD3dDevice))))
		{
			MLOG_ERROR("Failed to create D3D12 device, will try WARP.");
			return false;
		}
		else
		{
			Microsoft::WRL::ComPtr<IDXGIAdapter> pWarpAdapter;
			mDxgiFactory->EnumWarpAdapter(IID_PPV_ARGS(&pWarpAdapter));
			if (FAILED(::D3D12CreateDevice(pWarpAdapter.Get(), D3D_FEATURE_LEVEL_11_0, IID_PPV_ARGS(&mD3dDevice))))
			{
				MLOG_ERROR("Failed to create WARP device.");
				return false;
			}
		}
	}
	else
	{
		MLOG_ERROR("Failed to get DirectX adapter.");
		return false;
	}
	return true;
}

bool Win32DxInterface::createCommandQueue()
{
	D3D12_COMMAND_QUEUE_DESC queueDesc = {};
	queueDesc.Flags = D3D12_COMMAND_QUEUE_FLAG_NONE;
	queueDesc.Type = D3D12_COMMAND_LIST_TYPE_DIRECT;
	if (FAILED(mD3dDevice->CreateCommandQueue(&queueDesc, IID_PPV_ARGS(&mCommandQueue))))
	{
		MLOG_ERROR("Failed to create command queue.");
		return false;
	}
	return true;
}

bool Win32DxInterface::initializeD2d()
{
	if (!initializeD11on12())
	{
		return false;
	}

	D2D1_FACTORY_OPTIONS d2dFactoryOptions = {};
	D2D1CreateFactory(D2D1_FACTORY_TYPE_SINGLE_THREADED, __uuidof(ID2D1Factory3), &d2dFactoryOptions, &mD2dFactory);
	Microsoft::WRL::ComPtr<IDXGIDevice> dxgiDevice;
	mD3d11On12Device.As(&dxgiDevice);

	mD2dFactory->CreateDevice(dxgiDevice.Get(), &mD2dDevice);

	D2D1_DEVICE_CONTEXT_OPTIONS deviceOptions = D2D1_DEVICE_CONTEXT_OPTIONS_NONE;
	mD2dDevice->CreateDeviceContext(deviceOptions, &mD2dDeviceContext);

	return initializeDirectWrite();
}

bool Win32DxInterface::initializeD11on12()
{
	Microsoft::WRL::ComPtr<ID3D11Device> d3d11Device;
	UINT d3d11DeviceFlags = D3D11_CREATE_DEVICE_BGRA_SUPPORT;
	auto pCommandQueue = reinterpret_cast<IUnknown**>(mCommandQueue.GetAddressOf());
	if (FAILED(D3D11On12CreateDevice(mD3dDevice.Get(), d3d11DeviceFlags, nullptr, 0, pCommandQueue, 1, 0, &d3d11Device, &mD3d11DeviceContext, nullptr)))
	{
		MLOG_ERROR("Failed to create D3D11 on 12 Device");
		return false;
	}
	d3d11Device.As(&mD3d11On12Device);
	return true;
}

bool Win32DxInterface::initializeDirectWrite()
{
	return SUCCEEDED(::DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory), &mDWriteFactory));
}